package ch.inf3.horizoncut.cli;

import ch.inf3.horizoncut.data.BasicLayer;
import ch.inf3.horizoncut.data.DisplayCalculator;
import ch.inf3.horizoncut.data.Position;
import ch.inf3.horizoncut.data.Tile;
import ch.inf3.horizoncut.data.TileMap;
import ch.inf3.horizoncut.data.TileReader;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.NavigableMap;
import java.util.TreeMap;
import javax.imageio.ImageIO;
import net.e175.klaus.solarpositioning.AzimuthZenithAngle;
import net.e175.klaus.solarpositioning.DeltaT;
import net.e175.klaus.solarpositioning.SPA;
import net.e175.klaus.solarpositioning.SunriseTransitSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Cli2 {
    private static final Logger LOG = LoggerFactory.getLogger(Cli2.class);
    
    private static final int SLICE_DISTANCE = 50;
    
    public static void main(String[] args) throws Exception {
        
        long initStart = System.currentTimeMillis();
        Position eyePosition = new Position(46.28747, 7.44159);
        TileMap tileMap = new TileReader("/home/valdor/Downloads/SRTM1/").readHGTFiles(eyePosition, 2);

        final int width = 7200;
        final int height = 1000;

        Tile atOrigin = tileMap.getByLatLon(eyePosition.gridLat, eyePosition.gridLon);
        double eyeHeight = atOrigin.elevation(eyePosition.row, eyePosition.col) + 5;
        eyePosition.setHeight(eyeHeight);

        double minHorizAngle = 0;
        double maxHorizAngle = 359;

        DisplayCalculator dc = new DisplayCalculator(width, height, minHorizAngle, maxHorizAngle, eyePosition);

        int maxDistance = 100_000;
        int minDistance = 100;
        
        long initStop = System.currentTimeMillis();
        LOG.info("Initialized system for computations from {}@{}m with distances of {} to {} ({} ms)",
                eyePosition.getLatLon(), eyeHeight, minDistance, maxDistance, initStop - initStart);

        long layersStart = System.currentTimeMillis();
        final NavigableMap<Integer, BasicLayer> dLayers = new TreeMap<>();
        
        for (int d = maxDistance; d > minDistance; d -= SLICE_DISTANCE) {
            dLayers.put(d, BasicLayer.atDistance(d - SLICE_DISTANCE, d, tileMap, dc));
        }
        
        long layersEnd = System.currentTimeMillis();
        LOG.info("Computed {} layers ({} ms)", dLayers.size(), layersEnd - layersStart);
        
        long flattenStart = System.currentTimeMillis();
        
        BasicLayer horizonLayer = dLayers.values().stream()
                .reduce((a, b) -> a.mergeWith(b))
                .get();

        long flattenEnd = System.currentTimeMillis();
        LOG.info("Merged horizon from {} layers in {} ms - {} to {}", dLayers.size(),
                flattenEnd - flattenStart,
                horizonLayer.getNearDistance(), horizonLayer.getFarDistance());
        
        computeSunVisible(horizonLayer, eyePosition);

        long sunStart = System.currentTimeMillis();
        
        AzimuthZenithAngle[] sunPosition = new AzimuthZenithAngle[24];
        
        for (int i = 0; i < 24; i += 1) {
            ZonedDateTime dateTime = ZonedDateTime.now()
                    .withHour(i)
                    .withMinute(0);

            // replace SPA with Grena3 as needed
            AzimuthZenithAngle position = SPA.calculateSolarPosition(
                    dateTime,
                    eyePosition.getLatitude(), // latitude (degrees)
                    eyePosition.getLongitude(), // longitude (degrees)
                    eyePosition.getHeight(), // elevation (m)
                    DeltaT.estimate(dateTime.toLocalDate())); // delta T (s)
            
            sunPosition[i] = position;
        }
        
        ZonedDateTime dt = ZonedDateTime.now();
        
        SunriseTransitSet res = SPA.calculateSunriseTransitSet(dt,
                eyePosition.getLatitude(),
                eyePosition.getLongitude(),
                DeltaT.estimate(dt.toLocalDate()));
        
        LOG.info("Transits are {}", res);
        
        long sunEnd = System.currentTimeMillis();
        LOG.info("Computed sun position in {}ms", sunEnd - sunStart);
        
        long renderStart = System.currentTimeMillis();
        
        dc.setMinMaxVertAngle(80, 180);

        BufferedImage bufferedImg = new BufferedImage(width, height + 1, BufferedImage.TYPE_INT_RGB);
        Graphics2D g = (Graphics2D) bufferedImg.getGraphics();

        g.setColor(Color.RED);

        Polygon p = new Polygon();
        p.addPoint(0, height);
        for (int x = 0; x < width; x++) {
            double bearing = dc.getBearingAtX(x);
            double visibleAngle = horizonLayer.getAtBearing(bearing);
            
            int y = dc.getYAtAngle(visibleAngle);
            
            p.addPoint(x, y);
        }

        p.addPoint(width, height);
        g.drawPolygon(p);

        for (int i = 0; i < 360; i += 10) {
            if (i % 90 == 0) {
                g.setColor(Color.WHITE);
            } else {
                g.setColor(Color.GRAY);
            }
            
            int x = dc.getXAtBearing(i);

            g.drawLine(x, 0, x, height);
            g.drawString(i + "", x + 5, 20);
        }
        for (double i = dc.getMinHorizAngle(); i < dc.getMaxHorizAngle(); i += 5) {
            if (i % 90 == 0) {
                g.setColor(Color.WHITE);
            } else {
                g.setColor(Color.GRAY);
            }
            
            int y = dc.getYAtAngle(i);

            g.drawLine(0, y, width, y);
            g.drawString(i + "", 20, y + 5);
        }
        
        g.setColor(Color.YELLOW);
        for (int i = 0; i < 24; i++){
            AzimuthZenithAngle sp = sunPosition[i];
            int x = dc.getXAtBearing(sp.getAzimuth());
            int y = dc.getYAtAngle(180 - sp.getZenithAngle());
            
            g.fillOval(x - 10, y - 10, 20, 20);
            g.drawString(String.format("%02d:00", i), x - 40, y);
        }

        long renderEnd = System.currentTimeMillis();
        LOG.info("Rendered panorama ({} ms)", renderEnd - renderStart);

        g.dispose();

        try {
            String imageName = String.format("image-%05d.png", 1);
            File outputfile = new File("/home/valdor/", imageName);
            outputfile.createNewFile();

            ImageIO.write(bufferedImg, "png", outputfile);
        } catch (IOException ex) {
            LOG.error("Exception while writing image", ex);
        }
    }
    
    public static AzimuthZenithAngle computeSunAtTime(Position eyePosition, ZonedDateTime dateTime) {
        return SPA.calculateSolarPosition(
                    dateTime,
                    eyePosition.getLatitude(), // latitude (degrees)
                    eyePosition.getLongitude(), // longitude (degrees)
                    eyePosition.getHeight(), // elevation (m)
                    DeltaT.estimate(dateTime.toLocalDate())); // delta T (s)
    }
    
    public static ZonedDateTime computeSunVisible(BasicLayer dl, Position eyePosition, ZonedDateTime beforeDate, ZonedDateTime afterDate) {
        long rangeDiffSeconds = afterDate.toEpochSecond() - beforeDate.toEpochSecond();
        
        if (rangeDiffSeconds < 60) {
            return beforeDate;
        }
        
        ZonedDateTime middleDate = beforeDate.plusSeconds(rangeDiffSeconds / 2);
        AzimuthZenithAngle beforePosition = computeSunAtTime(eyePosition, beforeDate);
        AzimuthZenithAngle middlePosition = computeSunAtTime(eyePosition, middleDate);
        AzimuthZenithAngle afterPosition = computeSunAtTime(eyePosition, afterDate);

        boolean sunBeforeVisible = isSunVisible(dl, beforePosition);
        boolean sunMiddleVisible = isSunVisible(dl, middlePosition);
        boolean sunAfterVisible = isSunVisible(dl, afterPosition);
        
        if (sunBeforeVisible && sunAfterVisible) {
            throw new IllegalArgumentException("Sun is visible both before and after. Invalid dates");
        }
        
        if (sunBeforeVisible) {
            if (sunMiddleVisible) {
                return computeSunVisible(dl, eyePosition, middleDate, afterDate);
            } else {
                return computeSunVisible(dl, eyePosition, beforeDate, middleDate);
            }
        } else {
            if (sunMiddleVisible) {
                return computeSunVisible(dl, eyePosition, beforeDate, middleDate);
            } else {
                return computeSunVisible(dl, eyePosition, middleDate, afterDate);
            }
        }
    }
    
    public static void computeSunVisible(BasicLayer dl, Position eyePosition) {
        ZonedDateTime sunriseDate = computeSunVisible(dl, eyePosition,
                ZonedDateTime.now()
                        .withHour(3)
                        .withMinute(0)
                        .withSecond(0),
                ZonedDateTime.now()
                        .withHour(12)
                        .withMinute(0)
                        .withSecond(0));

        LOG.info("Found sunrise at {}", sunriseDate);
        ZonedDateTime sunsetDate = computeSunVisible(dl, eyePosition,
                ZonedDateTime.now()
                        .withHour(15)
                        .withMinute(0),
                ZonedDateTime.now()
                        .withHour(23)
                        .withMinute(0));

        LOG.info("Found sunset at {}", sunsetDate);
    }
    
    public static boolean isSunVisible(BasicLayer dl, AzimuthZenithAngle sp) {
        double sunBearing = sp.getAzimuth();
        double sunAngle = 180 - sp.getZenithAngle();
        double terrainAngle = dl.getAtBearing(sunBearing);

        return sunAngle >= terrainAngle;
    }
}
