package ch.inf3.horizoncut.data;

import java.util.Arrays;
import java.util.Collection;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class BasicLayer implements Comparable<BasicLayer>, Cloneable {

    private static final Logger LOG = LoggerFactory.getLogger(BasicLayer.class);

    protected final int nearDistance;
    protected final int farDistance;
    protected final double[] data;
    
    protected final DisplayCalculator dc;
    
    protected BasicLayer(int nearDistance, int farDistance, double[] data, DisplayCalculator dc) {
        this.nearDistance = nearDistance;
        this.farDistance = farDistance;
        this.data = data;
        this.dc = dc;
    }

    public static BasicLayer atDistance(int nearDistance, int farDistance, TileMap tileMap, DisplayCalculator dc) {

        double[] output = new double[dc.getWidth()];

        for (int x = 0; x < dc.getWidth(); x += 1) {
            double bearing = dc.getBearingAtX(x);
            Position target = Position.get(dc.getEyePosition(), Math.toRadians(bearing), nearDistance);

            Tile atPos = tileMap.getByLatLon(target.gridLat, target.gridLon);
            double elevation = atPos.elevation(target.row, target.col);

            int dMeters = dc.getEyePosition().distanceToMeters(target);
            double visibleAngle = dc.calculateVisibilityAngle(elevation, dMeters);

            output[x] = visibleAngle;
        }

        return new BasicLayer(nearDistance, farDistance, output, dc);
    }

    public int getNearDistance() {
        return nearDistance;
    }

    public int getFarDistance() {
        return farDistance;
    }
    
    public double getAtBearing(double bearing) {
        int x = dc.getXAtBearing(bearing);
        
        return data[x];
    }

    @Override
    public int compareTo(BasicLayer other) {
        return Integer.compare(this.nearDistance, other.nearDistance);
    }

    @Override
    protected BasicLayer clone() {
        double[] cloneData = Arrays.copyOf(data, data.length);

        return new BasicLayer(nearDistance, farDistance, cloneData, dc);
    }

    public BasicLayer mergeWith(BasicLayer other) {
        BasicLayer near = this.compareTo(other) <= 0 ? this : other;
        BasicLayer far = this.compareTo(other) > 0 ? this : other;

        double[] output = new double[near.data.length];

        int countNear = 0;
        int countFar = 0;

        for (int d = 0; d < near.data.length; d += 1) {
            double yNear = near.data[d];
            double yFar = far.data[d];

            if (yNear >= yFar) {
                output[d] = yNear;
                countNear += 1;
            } else {
                output[d] = yFar;
                countFar += 1;
            }
        }
        
        int tolerance = near.data.length / 100;

        int outputNearDistance = (countNear > tolerance) ? near.nearDistance : far.nearDistance;
        int outputFarDistance = (countFar > tolerance) ? far.farDistance : near.farDistance;
        
        LOG.debug("Merging {}-{} ({}) with {}-{} ({}) {}",
                near.nearDistance, near.farDistance, countNear,
                far.nearDistance, far.farDistance, countFar,
                tolerance);

        return new BasicLayer(outputNearDistance, outputFarDistance, output, dc);
    }

    /**
     * Returns true if the given layer is always lower than the current layer.
     *
     * @param other
     * @return
     */
    public boolean isLower(BasicLayer other) {
        for (int d = 0; d < this.data.length; d += 1) {
            double yThis = this.data[d];
            double yOther = other.data[d];

            if (yThis == 0 || yOther == 0) {
                continue;
            }

            if (yOther >= yThis) {
                return false;
            }
        }

        return true;
    }

    public double getMax() {
        return Arrays.stream(this.data)
                .boxed()
                .filter(e -> !e.isNaN())
                .max(Double::compareTo)
                .orElse(0.0);
    }

    public static boolean isVisible(Collection<? extends BasicLayer> sortedLayers, int distance, int x, double visibleAngle) {
        for (BasicLayer l : sortedLayers) {
            if (l.farDistance > (distance - 200)) {
                return true;
            }

            double c = l.data[x];

            if (c >= visibleAngle) {
                return false;
            }
        }

        return true;
    }

}
